(*
 * <COPYRIGHT>
 * Copyright 2005 Gerd Stolpmann
 *
 * <GPL>
 * This file is part of UMLMON.
 *
 * UMLMON is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * UMLMON is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with UMLMON; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * </>
 *)

(* $Id: umlmon.ml 209 2005-10-23 13:46:46Z gerd $ *)


(**********************************************************************)
(* Imports *)

open Umlmonctrl_aux
open Umlmon_types
open Umlmon_common

open Printf

(**********************************************************************)
(* Global state *)

let ues = Unixqueue.create_unix_event_system()

let vm = 
  { umlglobal = lazy(Umlmon_config.read_global_config());
    umlspec = lazy(failwith "UMLMON configuration not yet loaded");
    umlvm = `Inactive;
    md = None;
    ctrlsrv = [];
  }


(**********************************************************************)
(* Setup functions *)


let setup_vardir() =
  let global = Lazy.force vm.umlglobal in
  let spec = Lazy.force vm.umlspec in
  let uid = spec.umlpriv.priv_uid in
  let gid = spec.umlpriv.priv_gid in
  let umldir = Filename.concat global.umlvardir spec.umlid in
  let std_jaildir = Filename.concat umldir "jail" in
  let std_logdir = Filename.concat umldir "log" in

  List.iter
    (fun (dir, mode, uid, gid) -> 
       if not (Sys.file_exists dir) then (
	 Unix.mkdir dir mode;    (* modulo umask = 0o022 *)
	 Unix.chown dir uid gid;
       )
    )
    (
      [ global.umlvardir, 0o755, 0, 0;
        umldir, 0o755, 0, 0;
      ]
      @ (if std_jaildir = spec.umljaildir then
	   [ std_jaildir, 0o755, 0, 0;
	     (Filename.concat std_jaildir "disks"), 0o700, uid, gid;
	     (Filename.concat std_jaildir "files"), 0o700, uid, gid;
	     (Filename.concat std_jaildir "tmp"), 0o000, 0, 0;
	     (Filename.concat std_jaildir "proc"), 0o755, 0, 0;
	     (* "shared": is created in setup_shared *)
	     (* "dev": is created in setup_jail *)
	   ]
	 else [])
      @ (if std_logdir = spec.umllogdir then
	   [ std_logdir, 0o755, 0, 0 ]
	 else
	   [])
    )
;;

let setup_shared() =
  (* Establish the bind-mount to the shared directory if configured.
   * It is no problem to raise exceptions.
   *)
  let global = Lazy.force vm.umlglobal in
  let umlspec = Lazy.force vm.umlspec in
  let uid = umlspec.umlpriv.priv_uid in
  let gid = umlspec.umlpriv.priv_gid in
  let jd = umlspec.umljaildir in
  if jd <> "/" then (
    let shared = global.umlshared in
    if not(Sys.file_exists shared) then
      failwith ("Directory does not exist: " ^ shared);
    let jd_shared = Filename.concat jd "shared" in
    ( try
	Mount.mount_bind_dir shared jd_shared;
	monitor_log (sprintf "umlmon: Mounted %s" jd_shared)
      with
	| Mount.Already_mounted ->
	    monitor_log (sprintf
			   "umlmon: Something is already mounted at %s, not trying to mount the shared directory"
			   jd_shared);
	    monitor_log "umlmon: Continuing, in the hope this is a forgotten umount";
	| Mount.Mount_point_not_empty ->
	    failwith ("Not empty or not a directory: " ^ jd_shared)
	| Mount.Mount_error ->
	    failwith (sprintf
			"Cannot bind-mount %s at %s"
			shared
			jd_shared)
    );
  )


let clean_shared() =
  (* Remove the bind mount of the shared directory if existing. Avoid
   * exceptions.
   *)
  let global = Lazy.force vm.umlglobal in
  let umlspec = Lazy.force vm.umlspec in
  let jd = umlspec.umljaildir in
  let jd_shared = Filename.concat jd "shared" in
  if (Sys.file_exists jd_shared) then (
    ( try
	Mount.umount jd_shared;
	monitor_log (sprintf "umlmon: Umounted %s" jd_shared)
      with
	| Mount.Not_a_mount_point ->
	    ()   (* This may happen if no shared directory is configured *)
	| Mount.Mount_error ->
	    monitor_log (sprintf "umlmon: Cannot umount: %s" jd_shared)
    )
  )


let setup_jail() =
  (* Setup the chroot-based jail. It is allowed to raise [Failure], but
   * avoid other exceptions.
   *)
  let umlspec = Lazy.force vm.umlspec in
  let uid = umlspec.umlpriv.priv_uid in
  let gid = umlspec.umlpriv.priv_gid in
  let jd = umlspec.umljaildir in
  if jd <> "/" then (
    let jd_tmp = Filename.concat jd "tmp" in
    ( try
	Mount.mount_tmpfs (umlspec.umlmem + 2) jd_tmp;
	monitor_log (sprintf "umlmon: Mounted %s" jd_tmp)
      with
	| Mount.Already_mounted ->
	    monitor_log (sprintf
			   "umlmon: Something is already mounted at %s, not trying to mount the tmpfs directory"
				 jd_tmp);
	    monitor_log "umlmon: Continuing, in the hope this is a forgotten umount";
	| Mount.Mount_point_not_empty ->
	    failwith ("Not empty or not a directory: " ^ jd_tmp)
	| Mount.Mount_error ->
	    failwith (sprintf "Cannot mount tmpfs at %s" jd_tmp)
    );
    Unix.chown jd_tmp uid gid;

    let jd_proc = Filename.concat jd "proc" in
    if not(Sys.file_exists jd_proc) then (
      Unix.mkdir jd_proc 0o755;
    );

    if Sys.file_exists "/proc/mm" then (
      let jd_proc_mm = Filename.concat jd_proc "mm" in
      ( try
	  Mount.mount_bind_file "/proc/mm" jd_proc_mm;
	  monitor_log (sprintf "umlmon: Mounted %s" jd_proc_mm)
      with
	| Mount.Already_mounted ->
	    monitor_log (sprintf
			   "umlmon: Something is already mounted at %s, not trying to mount /proc/mm"
			   jd_proc_mm);
	    monitor_log "umlmon: Continuing, in the hope this is a forgotten umount";
	| Mount.Mount_point_not_empty ->
	    failwith ("Not empty or not a directory: " ^ jd_proc_mm)
	| Mount.Mount_error ->
	    failwith (sprintf "Cannot mount tmpfs at %s" jd_proc_mm)
      );
    );

    if Sys.file_exists "/proc/cpuinfo" then (
      Linux.File.copy "/proc/cpuinfo"  (Filename.concat jd_proc "cpuinfo")
    );

    let jd_dev = Filename.concat jd "dev" in
    if not(Sys.file_exists jd_dev) then (
      Unix.mkdir jd_dev 0o755;
      ignore(Sys.command(Printf.sprintf "mknod -m 666 %s/zero c 1 5" jd_dev));
      ignore(Sys.command(Printf.sprintf "mknod -m 666 %s/null c 1 3" jd_dev));
      ignore(Sys.command(Printf.sprintf "mknod -m 666 %s/rtc c 10 135" jd_dev));
      ignore(Sys.command(Printf.sprintf "mknod -m 666 %s/random c 1 8" jd_dev));
      ignore(Sys.command(Printf.sprintf "mknod -m 666 %s/urandom c 1 9" jd_dev));
      let jd_dev_net = Filename.concat jd_dev "net" in
      Unix.mkdir jd_dev_net 0o755;
      ignore(Sys.command(Printf.sprintf "mknod -m 660 %s/tun c 10 200" jd_dev_net));
      Unix.chown (Filename.concat jd_dev_net "tun") 0 gid
    )
  )


let clean_jail() =
  (* Remove the mounts in the jail directory. Avoid exceptions. *)
  let umlspec = Lazy.force vm.umlspec in
  let jd = umlspec.umljaildir in
  let jd_tmp = Filename.concat jd "tmp" in
  (* Check whether jd_tmp is a mount point. If not, umount the filesystem: *)
  ( try
      Mount.umount jd_tmp;
      monitor_log (sprintf "umlmon: Umounted %s" jd_tmp)
    with
      | Mount.Not_a_mount_point ->
	  ()
      | Mount.Mount_error ->
	  monitor_log (sprintf "umlmon: Cannot umount: %s" jd_tmp)
  );

  let jd_proc_mm = Filename.concat jd "proc/mm" in
  (* Check whether jd_proc_mm is a mount point. If so, umount the procfs filesystem: *)

  ( try
      Mount.umount jd_proc_mm;
      monitor_log (sprintf "umlmon: Umounted %s" jd_proc_mm)
    with
      | Mount.Not_a_mount_point ->
	  ()
      | Mount.Mount_error ->
	  monitor_log (sprintf "umlmon: Cannot umount: %s" jd_proc_mm)
  )
    

(**********************************************************************)
(* The "internal" RPC service *)

let proc_uml_statchange_running session pid reply =
  monitor_log "umlmon: got notification statchange_running";
  match vm.umlvm with
    | `Inactive ->
	reply ();
	failwith "Illegal state change from 'inactive' to 'running'"
    | `Finishing ->
	reply ();
	failwith "Illegal state change from 'finishing' to 'running'"
    | `Stopped actv ->
	actv.actv_vmpid <- Some (Int32.to_int pid);
	vm.umlvm <- `Running actv;
	reply ()
    | `Running actv ->
	actv.actv_vmpid <- Some (Int32.to_int pid);
	reply ()
    | `Starting actv ->
	actv.actv_vmpid <- Some (Int32.to_int pid);
	vm.umlvm <- `Running actv;
	reply ()
;;


let deactivate vm =
  match vm.umlvm with
    | `Inactive 
    | `Finishing ->
	(* Ignore... we are already down *)
	();
    | `Stopped actv ->
	vm.umlvm <- `Finishing;
	Umlmon_term.disconnect_terminals actv.actv_umlterms;
    | `Running actv ->
	vm.umlvm <- `Finishing;
	Umlmon_term.disconnect_terminals actv.actv_umlterms;
    | `Starting actv ->
	vm.umlvm <- `Finishing;
	Umlmon_term.disconnect_terminals actv.actv_umlterms


let proc_uml_statchange_exit session exit_code reply =
  monitor_log "umlmon: got notification statchange_exit";
  deactivate vm;
  reply ()
;;


let proc_uml_statchange_error session () reply =
  monitor_log "umlmon: got notification statchange_error";
  deactivate vm;
  reply ()
;;


(**********************************************************************)
(* UML startup *)

(* The UML monitor forks a child (the "monitor child") that performs:
 * - chroot to the jail
 * - drops privileges
 * - starts the kernel as child
 * - watches the kernel process, and sends the messages
 *     * Internal.V1.uml_statchange_running: when the kernel process has
 *       just been started
 *     * Internal.V1.uml_statchange_exit: when the kernel process has
 *       just exited.
 * - finally, waits until it is killed
 *
 * When the monitor child forks, the following things already have happened:
 * - Monitored terminals have been created (there is a [umlterm_activation list]).
 *   The file descriptors are shared with the monitor master process. The
 *   monitor child closes all descriptors after they have been used.
 * - The jail is completely set up
 *
 * The terminal "con0" is, if present, used for stdin/stdout of the kernel,
 * and it also becomes the controlling terminal.
 *)


let uml_startup actv_terms additional_kernel_args lock_id =
  let global = Lazy.force vm.umlglobal in
  let umlspec = Lazy.force vm.umlspec in
  let uid = umlspec.umlpriv.priv_uid in
  let gid = umlspec.umlpriv.priv_gid in
  
  let (fd0, fd1) = Unix.socketpair Unix.PF_UNIX Unix.SOCK_STREAM 0 in

  let (err0, err1) = Unix.pipe() in

  match Unix.fork() with
    | 0 ->
	(* --- Child process --- *)
	Unix.dup2 err1 Unix.stderr;

	monitor_log "umlmon [child]: start";

	(* Close all descriptors except stdin,stdout,stderr and fd1: *)
	let fd_keep = [ Unix.stdin; Unix.stdout; Unix.stderr; fd1 ] in
	for n = 0 to Unix_exts.sysconf_open_max() - 1 do
	  let fd = Unix_exts.file_descr_of_int n in
	  if not (List.mem fd fd_keep) then
	    try Unix.close fd with _ -> () 
	done;

	let ues = Unixqueue.create_unix_event_system() in
	let int_rpc =
	  Umlmonctrl_clnt.Internal.V1.create_client
	    ~esys:ues (Rpc_client.Descriptor fd1) Rpc.Tcp in

	let console = ref None in
	( try
	    Sys.set_signal Sys.sigpipe Sys.Signal_default;
	    (* This process becomes a session leader: *)
	    ignore(Unix.setsid());  
	       (* Works because the parent is the process group leader *)
	    (* Look up "con0": *)
	    let at_con0_opt =
	      try Some(List.find (fun at -> at.actv_termname = "con0") actv_terms)
	      with Not_found -> None in
	    (* Reassign stdin/stdout: *)
	    ( match at_con0_opt with
		| None ->
		    (* No con0: Just make stdin/stdout = /dev/null *)
		    Unix.close Unix.stdin;
		    ignore(Unix.openfile "/dev/null" [Unix.O_RDONLY] 0);
		    Unix.close Unix.stdout;
		    ignore(Unix.openfile "/dev/null" [Unix.O_WRONLY] 0);
		| Some at_con0 ->
		    (* Have con0: open the terminal for stdin/stdout. The terminal
		     * becomes the controlling terminal.
		     *)
		    ( match at_con0.actv_termstatus with
			| `Pty pty ->
			    Unix.close Unix.stdin;
			    ignore(Unix.openfile pty.ptyslavename [Unix.O_RDONLY] 0);
			    Unix.close Unix.stdout;
			    ignore(Unix.openfile pty.ptyslavename [Unix.O_WRONLY] 0);
			    console := Some Unix.stdout;
		    )
	    );
	    (* Open terminals: *)
	    let pty_fds = Hashtbl.create 20 in
	    List.iter 
	      (fun at ->
		 let `Pty { ptyslavename = slavename } = at.actv_termstatus in
		 let fd = Unix.openfile slavename [Unix.O_RDWR; Unix.O_NOCTTY] 0 in
		 Hashtbl.add pty_fds at.actv_termname fd
	      )
	      actv_terms;
	    (* Jail myself: *)
	    Unix.chroot umlspec.umljaildir;
	    Unix.chdir "/";
	    ( match umlspec.umlsched with
		| `Default sched ->
		    Linux.Process.set_sched_other sched.sched_def_prio
		| `RoundRobin sched ->
		    Linux.Process.set_sched_rr sched.sched_rr_prio
	    );
	    (* Drop privileges: *)
	    (* Cap.drop_privileges(); *)     (* Just to be sure no capability survives! *)
	    Linux.Privilege.initgroups umlspec.umlvmuser gid;
	    Unix.setgid gid;
	    Unix.setuid uid;
	    (* Start the kernel: *)
	    let kernel_fds =
	      [ Unix.stdin; Unix.stdout; Unix.stderr ]
	      @ (Hashtbl.fold (fun _ fd acc -> fd :: acc) pty_fds []) in
	    let kernel_args = 
	      umlspec.umlkernelargs
	      @ (List.map
		   (fun at ->
		      let fd = Hashtbl.find pty_fds at.actv_termname in
		      at.actv_termname ^ "=fd:" ^ 
		      (string_of_int (Unix_exts.int_of_file_descr fd))
		   )
		   actv_terms) 
	      @ ["uml_dir=/tmp/uml" ]
	      @ ["umid=" ^ umlspec.umlid]
	      @ ["mem=" ^ string_of_int 
		           (max 8
			      (min umlspec.umlmemlimit umlspec.umlmem)) ^ "M"]
	      @ additional_kernel_args
	    in
	    let kernel_env = Shell_sys.create_env() in
	    Shell_sys.set_env_var kernel_env "USER" umlspec.umlvmuser;
	    Shell_sys.set_env_var kernel_env "LOGNAME" umlspec.umlvmuser;
	    Shell_sys.set_env_var kernel_env "PATH" "";
	    Shell_sys.set_env_var kernel_env "HOME" "/";
	    let kernel_eng = 
	      new Shell_uq.call_engine
		~ignore_error_code:true
		[ Shell.cmd 
		    ~environment:kernel_env
		    ~descriptors:kernel_fds
		    ~assignments:[ Shell.assign ~src:Unix.stderr ~target:Unix.stdout ]
		    umlspec.umlkernel kernel_args ] 
		ues
	    in
	    let kernel_ji = kernel_eng # job_instance in
	    let kernel_proc = List.hd (Shell_sys.processes kernel_ji) in
	    monitor_log "umlmon [child]: executing kernel";
	    (* Close all descriptors not needed: *)
	    Hashtbl.iter
	      (fun _ fd -> Unix.close fd)
	      pty_fds;
	    (* Setup events: *)
	    Uq_engines.when_state
	      ~is_done:(fun _ ->
			  monitor_log "umlmon [child]: kernel finished";
			  let st = Shell_sys.status kernel_proc in
			  let exit_code =
			    match st with
			      | Unix.WEXITED n -> n
			      | Unix.WSIGNALED n -> 255 
			      | _ -> assert false in
			  Umlmonctrl_clnt.Internal.V1.uml_statchange_exit'async
		             int_rpc
		             (Int32.of_int exit_code)
		             (fun reply -> ())
		       )
	      ~is_error:(function 
			   | Shell.Subprocess_error(_) ->
			       monitor_log "umlmon [child]: kernel signal";
			       Umlmonctrl_clnt.Internal.V1.uml_statchange_error'async
				 int_rpc
				 ()
				 (fun reply -> ())
			   | err ->
			       raise err)
	      kernel_eng;
	    (* The kernel is running. Tell the main process this: *)
	    Umlmonctrl_clnt.Internal.V1.uml_statchange_running'async
	      int_rpc
	      (Int32.of_int (Shell_sys.process_id kernel_proc))
	      (fun reply -> ());
	    (* Event loop *)
	    let cont = ref true in
	    while !cont do
	      try
		Unixqueue.run ues;
		cont := false
	      with
		| Sys.Break -> cont := false
		| err ->
		    monitor_log ("umlmon [child]: Caught exception: " ^ 
				   Printexc.to_string err ^ " - continuing")
	    done;

	  with
	      err ->
		monitor_log ("umlmon [child]: Caught exception: " ^ 
			       Printexc.to_string err ^ " - exiting");
	);
	monitor_log "umlmon [child]: finished";
	( match !console with
	    | Some fd -> 
		(* Just to be sure everything is transmitted *)
		Unix.tcdrain fd; 
		Unix.close Unix.stdin;
		Unix.close Unix.stdout;
		Unix.sleep 1
	    | None -> ()
	);

	exit 0
	
    | n ->
	(* --- Master process --- *)
	Unix.close fd1;
	Unix.close err1;

	(* Copy [err0] to [stderr]: *)
	let err_ch = new Uq_engines.output_async_descr 
		       ~dst:Unix.stderr ~close_dst:false ues in
	let _ = new Uq_engines.receiver ~src:err0 ~dst:err_ch ues in
	(* err0 will be closed when EOF is found *)

	(* Create the RPC server listening for messages of the child: *)
	let monsrv =
	  Umlmonctrl_srv.Internal.V1.create_async_server 
	    ~proc_ping:(fun _ _ reply -> reply())
	    ~proc_uml_statchange_running
	    ~proc_uml_statchange_exit
	    ~proc_uml_statchange_error
	    (Rpc_server.Descriptor fd0)
	    Rpc.Tcp
	    Rpc.BiPipe
	    ues in

	(* Arrange that fd0 is closed when the child process finishes: *)
	let e = new Uq_engines.poll_process_engine ~pid:n ues in
	Uq_engines.when_state 
	  ~is_done:(fun _ -> 
		      try
			vm.umlvm <- `Inactive;
			monitor_log "umlmon: child finished";
			Umlmon_shared.unlock_disks global umlspec lock_id;
			Rpc_server.stop_server monsrv;
			Unix.close fd0;
			Umlmon_term.shutdown_terminals actv_terms;
			clean_jail();
			Umlmon_net.clean_networking vm;
		      with
			  err ->
			    monitor_log ("umlmon: error in cleanup: " ^
					 Printexc.to_string err)
		   )
	  ~is_error:(fun _ -> 
		      try
			vm.umlvm <- `Inactive;
			monitor_log "umlmon: child error";
			Umlmon_shared.unlock_disks global umlspec lock_id;
			Rpc_server.stop_server monsrv;
			Unix.close fd0;
			Umlmon_term.shutdown_terminals actv_terms;
			clean_jail();
			Umlmon_net.clean_networking vm;
		      with
			  err ->
			    monitor_log ("umlmon: error in cleanup: " ^
					 Printexc.to_string err)
		    )
	  e;

	{ actv_vmpid = None;    (* PID of the VM is not yet known *)
	  actv_umlterms = actv_terms;
	  actv_monpid = n;
	  actv_monfd = fd0;
	  actv_monsrv = monsrv;
	  actv_lock = lock_id;
	}
;;

(**********************************************************************)
(* RPC procedures of the "external" service *)

let proc_ping session () reply =
  reply ()
;;

let proc_ensure_id session expected_id reply =
  let umlspec = Lazy.force vm.umlspec in
  if umlspec.umlid <> expected_id then (
    Rpc_server.stop_connection 
      (Rpc_server.get_server session)
      (Rpc_server.get_connection_id session)
  )
  else
    reply ()
;;

let proc_minor_protocol_version session () reply =
  reply (Rtypes.int32_of_uint4 Umlmonctrl_aux.minor_ctrl_version)
;;

let proc_available_extensions session () reply =
  reply [| "sync_disks" |]
;;

let proc_get_id = 
  protect 
    (fun session () reply ->
       Umlmon_auth.ensure_authenticated session;
       let umlspec = Lazy.force vm.umlspec in
       reply umlspec.umlid
    )
;;

let proc_get_status =
  protect
    (fun session () reply ->
       Umlmon_auth.ensure_authenticated session;
       match vm.umlvm with
	 | `Inactive  -> reply `ctrl_inactive
	 | `Running _ -> reply `ctrl_running
	 | `Finishing -> reply `ctrl_running   (* Counts as "Running" for observers *)
	 | `Stopped _ -> reply `ctrl_stopped
	 | `Starting _ -> reply `ctrl_starting
    )
;;

let proc_get_monitored_terminals = 
  protect
    (fun session () reply ->
       Umlmon_auth.ensure_authenticated session;
       let actv_terms =
	 match vm.umlvm with
	   | `Inactive -> []
	   | `Finishing -> []
	   | `Running vm -> vm.actv_umlterms
	   | `Stopped vm -> vm.actv_umlterms 
	   | `Starting vm -> vm.actv_umlterms 
       in
       let tarray =
	 Array.of_list
	   (List.map
	      (fun at ->
		 let td =
		   match at.actv_termstatus with
		     | `Pty pty ->
			 `ctrl_pty
			   { ctrl_ptsfile = Some pty.ptyslavename;
			     ctrl_logfile = pty.ptylogname;
			     ctrl_connected = (match pty.ptystate with
						 | `Connected _ -> true
						 | _ -> false);
			   }
		 in
		 { ctrl_termname = at.actv_termname;
		   ctrl_termdef = td;
		   ctrl_extension = [| |];
		 }
	      )
	      actv_terms
	   ) in
       reply tarray
    )
;;


let read_lines file start length =
  if not (Sys.file_exists file) then 
    []
  else (
    let f = open_in file in
    let lines = ref [] in
    try
      let n = ref 0l in
      while !n < start do
	ignore(input_line f);
	n := Int32.succ !n
      done;
      for k = 1 to length do
	lines := (input_line f) :: !lines
      done;
      close_in f;
      List.rev !lines
    with
      | End_of_file ->
	  close_in f;
	  List.rev !lines
      | err ->
	  close_in f;
	  raise err
  )
;;


let count_lines file =
  if not (Sys.file_exists file) then 
    0l
  else (
    let f = open_in file in
    let n = ref 0l in
    try
      while true do
	ignore(input_line f);
	n := Int32.succ !n;
	if !n < 0l then failwith "Overflow";
      done;
      assert false
    with
      | End_of_file ->
	  close_in f;
	  !n
      | err ->
	  close_in f;
	  raise err
  )
;;


let proc_logfile_read =
  protect 
    (fun session (filecode,startline,numlines) reply ->
       Umlmon_auth.ensure_authenticated session;
       let spec = Lazy.force vm.umlspec in
       try
	 if numlines < 1l then failwith "Bad number of lines";
	 if numlines > 1000l then failwith "Bad number of lines";
	 let file =
	   if filecode = "monitor" then
	     match spec.umlmonlog with
	       | None -> raise Not_found
	       | Some f -> f
	   else
	     let term = List.find (fun ut -> ut.termname = filecode) spec.umlterms in
	     match term.termdef with
	       | `Pty (Some f) -> f
	       | _ -> raise Not_found
	 in
	 if startline >= 0l then (
	   let lines = read_lines file startline (Int32.to_int numlines) in
	   reply (`code_ok { line_start = startline; lines = Array.of_list lines })
	 )
	 else (
	   let n = count_lines file in
	   let startline = max (Int32.add n startline) 0l in
	   let lines = read_lines file startline (Int32.to_int numlines) in
	   reply (`code_ok { line_start = startline; lines = Array.of_list lines })
	 )
       with
	 | Not_found ->
	     reply (`code_error "No such log file")
	 | Failure msg ->
	     reply (`code_error msg)
	 | Sys_error msg ->
	     reply (`code_error msg)
    )
;;


let proc_announced_kernels = 
  protect 
    (fun session () reply ->
       let global = Lazy.force vm.umlglobal in
       Umlmon_auth.ensure_authenticated session;
       let kernel_files = Umlmon_shared.get_kernels global in
       let kernels =
	 Array.of_list
	   (List.map
	      (fun (Jailfile kf) ->
		 { kernel_file = kf;
		   kernel_name = Filename.basename kf
		 }
	      )
	      kernel_files) in
       reply kernels
    )
;;


let proc_start =
  protect 
    (fun session args reply ->
       Umlmon_auth.ensure_authenticated session;
       match vm.umlvm with
	 | `Inactive ->
	     (* The UML is inactive, so start it now:
	      * - Allocate all terminals
	      * - Set up the jail
	      * - Fork the monitor child
	      *)
	     ( let err_actions = ref [] in
	       try
		 if Umlmon_disk.operation_in_progress vm then
		   failwith "Cannot start VM while disk operations are in progress";
		 (* Reread configuration: *)
		 let id = (Lazy.force vm.umlspec).umlid in
		 let spec = Umlmon_config.read_vm_config id in
		 vm.umlspec <- lazy spec;
		 reopen_stderr spec;
		 let global = Lazy.force vm.umlglobal in
		 let lock_id = Umlmon_disk.lock_disks_at_startup vm in
		               (* or Lock_failure *)
		 err_actions := (fun () -> 
				   Umlmon_shared.unlock_disks
				     global spec lock_id) :: !err_actions;
		 let disk_args = Umlmon_disk.kernel_arguments vm in
		 setup_jail();
		 err_actions :=  clean_jail :: !err_actions;
		 let net_setup = Umlmon_net.setup_networking vm in
		 err_actions := (fun () -> Umlmon_net.clean_networking vm) 
		                :: !err_actions;
		 let net_args = Umlmon_net.kernel_arguments net_setup in
		 let actv_terms = 
		   Umlmon_term.create_terminals ues spec vm.umlvm in
		 err_actions := (fun () -> 
				   Umlmon_term.shutdown_terminals
				     actv_terms) :: !err_actions;
		 let actv_rec =
		   uml_startup 
		     actv_terms 
		     (disk_args @ net_args @ Array.to_list args)
		     lock_id
		 in
		 vm.umlvm <- `Starting actv_rec;
		 reply `code_in_progress
	       with
		 | Failure msg ->
		     monitor_log ("umlmon: Cannot start VM: " ^ msg);
		     reply (`code_error msg);
		     List.iter (fun f -> f()) !err_actions;

		 | Lock_failure r ->
		     let msg = "Lock failure: " ^ r in
		     monitor_log ("umlmon: Cannot start VM: " ^ msg);
		     reply (`code_error msg);
		     List.iter (fun f -> f()) !err_actions;

		 | err ->
		     monitor_log ("umlmon: Cannot start VM: " ^
				    Printexc.to_string err);
		     reply (`code_error (Printexc.to_string err));
		     List.iter (fun f -> f()) !err_actions;
	     )

	 | _ ->
	     (* Already started. Do nothing *)
	     reply `code_ok
    )
;;

let proc_kill =
  protect
    (fun session () reply ->
       Umlmon_auth.ensure_authenticated session;
       try
	 let actv =
	   match vm.umlvm with
	     | `Inactive  -> raise Not_found
	     | `Finishing -> raise Not_found
	     | `Running a -> a
	     | `Stopped a -> a
	     | `Starting a -> raise Not_found in
	 match actv.actv_vmpid with
	   | None -> raise Not_found
	   | Some pid -> 
	       Unix.kill pid Sys.sigterm;
	       reply `code_ok
       with
	   Not_found ->
	     reply (`code_error "No process to kill")
    )
;;

let proc_rotate =
  protect
    (fun session () reply ->
       Umlmon_auth.ensure_authenticated session;
       try
	 let umlspec = Lazy.force vm.umlspec in
	 reopen_stderr umlspec;
	 ( match vm.umlvm with
	     | `Inactive ->  
		 reply `code_ok
	     | `Finishing ->  
		 reply `code_ok
	     | `Stopped actv -> 
		 Umlmon_term.rotate_terminals umlspec actv.actv_umlterms;
		 reply `code_ok
	     | `Running actv ->
		 Umlmon_term.rotate_terminals umlspec actv.actv_umlterms;
		 reply `code_ok
	     | `Starting actv ->
		 Umlmon_term.rotate_terminals umlspec actv.actv_umlterms;
		 reply `code_ok
	 )
       with
	   err ->
	     monitor_log ("umlmon: rotate: " ^ Printexc.to_string err);
	     reply (`code_error (Printexc.to_string err))
    )
;;


let proc_get_ssl =
  Umlmon_mconsole.proc_get_ssl vm ;;


let proc_get_con =
  Umlmon_mconsole.proc_get_con vm ;;


let proc_get_ubd =
  Umlmon_mconsole.proc_get_ubd vm ;;


let proc_get_eth =
  Umlmon_mconsole.proc_get_eth vm ;;


let proc_stop =
  Umlmon_mconsole.proc_stop vm ;;


let proc_continue =
  Umlmon_mconsole.proc_continue vm ;;


let proc_halt =
  Umlmon_mconsole.proc_halt vm ;;


let proc_ctrlaltdel =
  Umlmon_mconsole.proc_ctrlaltdel vm ;;


let proc_sysrq =
  Umlmon_mconsole.proc_sysrq vm ;;


let proc_connect =
  Umlmon_term.proc_connect ues vm ;;


let proc_read =
  Umlmon_term.proc_read ues vm ;;


let proc_write =
  Umlmon_term.proc_write ues vm ;;


let proc_disconnect =
  Umlmon_term.proc_disconnect ues vm ;;


let proc_is_connected =
  Umlmon_term.proc_is_connected ues vm ;;


let proc_te_init =
  Umlmon_term.proc_te_init ues vm ;;


let proc_te_read =
  Umlmon_term.proc_te_read ues vm ;;


let proc_te_write_keys =
  Umlmon_term.proc_te_write_keys ues vm ;;


let proc_te_state =
  Umlmon_term.proc_te_state ues vm ;;


let proc_te_reset =
  Umlmon_term.proc_te_reset ues vm ;;


let proc_disk_files = 
  Umlmon_disk.proc_disk_files ues vm ;;


let proc_disk_create = 
  Umlmon_disk.proc_disk_create ues vm ;;


let proc_disk_create_cow = 
  Umlmon_disk.proc_disk_create_cow ues vm ;;


let proc_disk_delete = 
  Umlmon_disk.proc_disk_delete ues vm ;;


let proc_disk_copy = 
  Umlmon_disk.proc_disk_copy ues vm ;;


let proc_disk_resize = 
  Umlmon_disk.proc_disk_resize ues vm ;;


let proc_disk_options =
  Umlmon_disk.proc_disk_options ues vm ;;


let proc_config_get = 
  Umlmon_config.proc_config_get vm ;;


let proc_config_set = 
  Umlmon_config.proc_config_set vm ;;


let proc_config_hot_add = 
  Umlmon_config.proc_config_hot_add vm ;;


let proc_config_hot_remove = 
  Umlmon_config.proc_config_hot_remove vm ;;


let proc_auth_get_challenge = 
  Umlmon_auth.proc_auth_get_challenge vm ;;


let proc_auth_pass_credentials = 
  Umlmon_auth.proc_auth_pass_credentials ues vm ;;


let proc_auth_reset = 
  Umlmon_auth.proc_auth_reset ues vm ;;


let proc_auth_change_password =
  Umlmon_auth.proc_auth_change_password ~write_vm_config:Umlmon_config.write_vm_config vm
;;


let proc_site_command = 
  protect
    (fun session args reply ->
       Umlmon_auth.ensure_authenticated session;
       let global = Lazy.force vm.umlglobal in
       let spec = Lazy.force vm.umlspec in
       try
	 match global.umlsitecmd with
	   | None ->
	       failwith "No site command configured"
	   | Some cmd ->
	       let status = 
		 match vm.umlvm with
		   | `Inactive  -> "INACTIVE"
		   | `Finishing -> "RUNNING"  (* this state is hidden *)
		   | `Running _ -> "RUNNING"
		   | `Stopped _ -> "STOPPED"
		   | `Starting _ -> "STARTING" in
	       let environment = Shell_sys.current_env() in
	       Shell_sys.set_env_var environment "UMLMON_ID" spec.umlid;
	       Shell_sys.set_env_var environment "UMLMON_VARDIR" global.umlvardir;
	       Shell_sys.set_env_var environment "UMLMON_LOGDIR" spec.umllogdir;
	       Shell_sys.set_env_var environment "UMLMON_JAILDIR" spec.umljaildir;
	       (* Shell_sys.set_env_var environment "UMLMON_SHARED" XXX; *)
	       Shell_sys.set_env_var environment "UMLMON_STATUS" status;
	       Shell_sys.set_env_var environment "UMLMON_USER" spec.umlvmuser;
	       let buf = Buffer.create 1000 in
	       let eng = 
		 new Shell_uq.call_engine 
		   ~ignore_error_code:true
		   ~stdin:Shell.from_dev_null
		   ~stdout:(Shell.to_buffer buf)
		   [ Shell.command 
		       ~arguments:args
		       ~environment
		       ~assignments:[ Shell.assign 
					~src:Shell.stderr ~target:Shell.stdout ]
		       cmd
		   ] 
		   ues in
	       let site_ji = eng # job_instance in
	       let site_proc = List.hd (Shell_sys.processes site_ji) in
	       Uq_engines.when_state
		 ~is_done:(fun _ ->
			     match Shell_sys.status site_proc with
			       | Unix.WEXITED 0 ->
				   reply(`code_ok(Buffer.contents buf))
			       | Unix.WEXITED n ->
				   reply(`code_error(Buffer.contents buf))
			       | Unix.WSIGNALED n ->
				   reply(`code_error("Site cmd terminated by signal"))
			       | _ ->
				   assert false
			  )
		 ~is_error:(fun e ->
			      match e with
				| Shell.Subprocess_error _ ->
				    reply(`code_error(Buffer.contents buf))
				| _ ->
				    reply(`code_error(Printexc.to_string e))
			   )
		 eng
       with
	 | Failure msg ->
	     reply (`code_error msg)
	 | Sys_error msg ->
	     reply (`code_error msg)
	 | Shell_sys.Executable_not_found msg ->
	     reply (`code_error ("Not found: " ^ msg))
    )
;;


let proc_port = 
  protect
    (fun session _ reply ->
       Umlmon_auth.ensure_authenticated session;
       try
	 let srv =
	   List.find
	     (fun srv ->
		match Rpc_server.get_main_socket_name srv with
		  | Unix.ADDR_INET _ -> true
		  | _ -> false
	     )
	     vm.ctrlsrv in
	 match Rpc_server.get_main_socket_name srv with
	   | Unix.ADDR_INET(_,port) ->
	       reply (Some (Int32.of_int port))
	   | _ ->
	       assert false
       with
	   Not_found ->
	     reply None
    )
;;


(**********************************************************************)
(* Ctrl server *)

module Conn = struct
  type t = Rpc_server.connection_id
  let compare : t -> t -> int = Pervasives.compare
end


module ConnSet = Set.Make(Conn)


let create_ctrl_server_for =
  Umlmonctrl_srv.Ctrl.V1.create_async_server
    ~proc_ping
    ~proc_ensure_id
    ~proc_minor_protocol_version
    ~proc_available_extensions
    ~proc_get_id
    ~proc_get_status
    ~proc_get_monitored_terminals
    ~proc_get_ssl
    ~proc_get_con
    ~proc_get_ubd
    ~proc_get_eth
    ~proc_port
    ~proc_logfile_read
    ~proc_announced_kernels
    ~proc_start
    ~proc_stop
    ~proc_continue
    ~proc_halt
    ~proc_ctrlaltdel
    ~proc_kill
    ~proc_rotate
    ~proc_sysrq
    ~proc_connect
    ~proc_read 
    ~proc_write
    ~proc_disconnect
    ~proc_is_connected
    ~proc_te_init
    ~proc_te_read
    ~proc_te_write_keys
    ~proc_te_state
    ~proc_te_reset
    ~proc_disk_files
    ~proc_disk_create
    ~proc_disk_create_cow
    ~proc_disk_delete
    ~proc_disk_copy
    ~proc_disk_resize
    ~proc_disk_options
    ~proc_config_get
    ~proc_config_set
    ~proc_config_hot_add
    ~proc_config_hot_remove
    ~proc_auth_get_challenge
    ~proc_auth_pass_credentials
    ~proc_auth_change_password
    ~proc_auth_reset
    ~proc_site_command


let max_conns = 100
  (* maximum number of parallel TCP connections *)
    
let create_ctrl_server() =
  let global = Lazy.force vm.umlglobal in
  let umlspec = Lazy.force vm.umlspec in

  (* Create Unix Domain server: *)
  let srv_local =
    let sockname = 
      Filename.concat (Filename.concat global.umlvardir umlspec.umlid) "ctrl" in
    (try Sys.remove sockname with _ -> ());
    let srv =
      create_ctrl_server_for
	(Rpc_server.Unix sockname)
	Rpc.Tcp
	Rpc.Socket
	ues in
    Unix.chmod sockname 0o666;
    Rpc_server.set_onclose_action
      srv (fun conn_id -> 
	     Umlmon_auth.forget_connection conn_id);
    srv
  in

  (* Create TCP server: *)
  let srv_tcp =
    if global.umltcp then (
      let srv =
	create_ctrl_server_for
	  (Rpc_server.Internet(Unix.inet_addr_any,0))
	  Rpc.Tcp
	  Rpc.Socket
	  ues in
      let n_conns = ref 0 in
      let known_conns = ref ConnSet.empty in
      Rpc_server.set_session_filter_2
	srv (fun _ conn_id -> 
	       if not (ConnSet.mem conn_id !known_conns) then (
		 known_conns := ConnSet.add conn_id !known_conns;
		 incr n_conns;
	       );
	       if !n_conns > max_conns || Umlmon_auth.known_as_failed conn_id then
		 `Deny 
	       else 
		 `Accept_limit_length(65536, `Deny));
      Rpc_server.set_onclose_action
	srv (fun conn_id -> 
	       decr n_conns;
	       known_conns := ConnSet.remove conn_id !known_conns;
	       Umlmon_auth.forget_connection conn_id);
      [ srv ]
    )
    else
      []
  in

  vm.ctrlsrv <- [ srv_local ] @ srv_tcp

(**********************************************************************)
(* Main *)

let version() =
  printf "UMLMON Release %s Revision %s\n" Instconf.release Instconf.revision;
  printf "Copyright 2005 by Informatikbuero Gerd Stolpmann.\n\n";
  printf "UMLMON is free software, and can be distributed under the terms\n";
  printf "of the GNU General Public License.\n\n";
  printf "Commercial support available, please contact gerd@gerd-stolpmann.de\n";
  flush stdout;
  exit 0
;;

let main() =
  let umlid = ref None in
  let fg = ref false in

  Arg.parse
    (Arg.align
       [ "-fg", Arg.Set fg, " Run in the foreground";
	"-version", Arg.Unit version, " Print version number and exit";
       ])
    (fun s ->
       match !umlid with
	 | None -> umlid := Some s
	 | Some _ -> raise(Arg.Bad "Only one VM can be controlled by this monitor!"))
    "usage: umlmon [options] <umlid>";

  let id =
    match !umlid with
      | None ->
	  failwith "Name of the VM is missing!";
      | Some id ->
	  id
  in

  if id = "directory" then
    failwith "Reserved name: directory";

  (* Check for syntax errors: *)
  ignore(Lazy.force vm.umlglobal);

  let spec = Umlmon_config.read_vm_config id in
  vm.umlspec <- lazy spec;

  if spec.umlstart = `Off then
    failwith "This VM instance is administratively disabled!";

  if not (Sys.file_exists "/var/run/umlmon") then (
    Unix.mkdir "/var/run/umlmon" 0o755
  );

  let global = Lazy.force vm.umlglobal in

  setup_vardir();
  reopen_stderr spec;
  Umlmon_shared.setup_shared global;  (* Create "shared" *)
  setup_shared();  (* Set up the "shared" mount in the jail *)
  Umlmon_shared.unlock_all_disks global spec;
  Umlmon_shared.release_all_hwaddr global spec;
  create_ctrl_server();
  daemon ~wait:!fg
    (Filename.concat "/var/run/umlmon" (id ^ ".pid"))
    (fun () ->
       monitor_log "umlmon: Startup";
       let cont = ref true in
       while !cont do
	 try
	   Unixqueue.run ues;
	   cont := false
	 with
	   | Sys.Break -> cont := false
	   | Auth_failed _ ->
	       monitor_log "umlmon: Failed authentication";
	   | err ->
	       monitor_log ("umlmon: Caught exception: " ^ Printexc.to_string err);
       done;
       clean_shared()
    )
;;

try
  Sys.set_signal Sys.sigpipe Sys.Signal_ignore;
  Unix.putenv "PATH" (Filename.concat Instconf.prefix "bin" ^
		   ":/bin:/sbin:/usr/bin:/usr/sbin");
  ignore(Unix.umask 0o22);
  main();
with
  | Failure msg ->
      prerr_endline ("umlmon: " ^ msg); exit 1
  | Sys_error msg ->
      prerr_endline ("umlmon: " ^ msg); exit 1
  | Unix.Unix_error(_,_,_) as e ->
      prerr_endline ("umlmon: " ^ Umlmon_common.string_of_unix_error e); exit 1
  | err ->
      prerr_endline ("umlmon: Caught exception " ^ Printexc.to_string err); exit 1
;;
