(*
 * <COPYRIGHT>
 * Copyright 2005 Gerd Stolpmann
 *
 * <GPL>
 * This file is part of UMLMON.
 *
 * UMLMON is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * UMLMON is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with UMLMON; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * </>
 *)

(* $Id: umlmon_config.ml 209 2005-10-23 13:46:46Z gerd $ *)

open Umlmon_types
open Umlmon_common
open Umlmonctrl_aux
open Cryptokit
open Printf


let default_sync_disks = [ "all" ];;

let check_etc_umlmon() =
  let st = Unix.stat "/etc/umlmon" in
  if st.Unix.st_uid <> 0 then
    failwith "/etc/umlmon: not owned by root";
  if st.Unix.st_perm <> 0o600 then
    failwith "/etc/umlmon: must have mode 600";
;;


let parse_password s =
  try
    if String.length s <> 32 then raise Not_found;
    for k = 0 to String.length s - 1 do
      match s.[k] with
	| '0'..'9' -> ()
	| 'a'..'f' -> ()
	| 'A'..'F' -> ()
	| _ -> raise Not_found
    done;
    (* [s] is a hex string. Take it as the MD5 sum of the password! *)
    let decoder = Hexa.decode() in
    transform_string decoder s
  with
      Not_found ->
	(* [s] is the unprotected password *)
	Digest.string s
;;


let print_password s =
  let encoder = Hexa.encode() in
  transform_string encoder s
;;


let fun_style_re =
  Str.regexp "^[ \t]*\\([a-zA-Z0-9_]+\\)[ \t]*(\\(.*\\))[ \t]*$";;

let fun_style_delim_re =
  Str.regexp ",";;

let fun_style_eq_re =
  Str.regexp "^[ \t]*\\([a-zA-Z0-9_]+\\)[ \t]*=[ \t]*\\(.*\\)[ \t]*$";;


exception Fun_style_not_recognized

let parse_fun_style ?scheme ?(raise_not_recognized=false) s =
  (* Parses an option with the format:
   *   fun_name(name1 = value1, name2 = value2, ... )
   * Returns the result as
   *   (fun_name, [name1, value1; name2, value2; ...])
   *
   * scheme: Declares which names are allowed:
   *    [ fun_name1, [name1; name2; ...];
   *      fun_name2, [name1; name2; ...]; ... ]
   * Names are tagged with `Opt (optional) or `Req (required).
   *)
  if Str.string_match fun_style_re s 0 then (
    let name = Str.matched_group 1 s in
    let list_s = Str.matched_group 2 s in
    let list_l = Str.split_delim fun_style_delim_re list_s in
    let list_eq =
      List.map
	(fun t ->
	   if Str.string_match fun_style_eq_re t 0 then (
	     let n = Str.matched_group 1 t in
	     let v = Str.matched_group 2 t in
	     (n,v)
	   )
	   else
	     failwith("Syntax error in: " ^ s)
	)
	list_l in
    ( match scheme with
	| None -> ()   (* No check *)
	| Some sch ->
	    ( try 
		let sch_params = List.assoc name sch in
		let names = 
		  List.map (function `Opt n -> n | `Req n -> n) sch_params in
		List.iter
		  (fun (n,_) ->
		     if not (List.mem n names) then
		       failwith("Unknown parameter '" ^ n ^ "' used in: " ^ s)
		  )
		  list_eq;
		List.iter
		  (function
		     | `Opt _ -> ()
		     | `Req n ->
			 if not (List.mem_assoc n list_eq) then
			   failwith("Required parameter '" ^ n ^ "' is missing in: "  ^ s)
		  )
		  sch_params;
	      with
		| Not_found ->
		    failwith("Unknown function '" ^ name ^ "' used in: " ^ s)
	    )
    );

    (name, list_eq)
  )
  else
    if raise_not_recognized then
      raise Fun_style_not_recognized
    else
      failwith("Syntax error in: " ^ s)
;;


let parse_device s =
  (* Syntax:
   * device_file(name = <name>, target_file = /dev/<target>, group = <group>,
   *     identity = ignore|check|update)
   *)
  let (_, params) =
    parse_fun_style
      ~scheme:[ "device_file", 
		[ `Req "name"; `Req "target_file"; `Opt "group"; 
		  `Opt "identity" ] ]
      s in
  let name = List.assoc "name" params in
  if String.contains name '/' then
    failwith ("Name must not contain '/', found in: " ^ s);
  if not (Filename.check_suffix name ".dev") then
    failwith ("Name must have extension .dev, found in: " ^ s);
  let target = List.assoc "target_file" params in
  let (group, perms) = 
    try 
      let gname = List.assoc "group" params in
      let gid =
	try (Unix.getgrnam gname).Unix.gr_gid
	with Not_found ->
	  failwith ("Unknown group found in: " ^ s) in
      (gid, 0o660)
    with
      | Not_found -> (0, 0o600) in  (* Default: root group *)
  (* CHECK: Without "group" parameter, the VMs can effectively not access
   * the device.
   *)
  let identity = 
    try
      match List.assoc "identity" params with
	| "ignore" -> `Ignore
	| "check" -> `Check
	| "update" -> `Update
	| _ -> failwith("Unknown identity keyword found in: " ^ s)
    with
	Not_found -> `Check
  in

  { dev_name = name;
    dev_target = target;
    dev_create_group = group;
    dev_create_perm = perms;
    dev_identity = identity;
  }
;;


let parse_scheduler s =
  (* Syntax:
   * sched_default(prio = <n>)
   * sched_rr(prio = <n>)
   *)
  let (schedname, params) =
    parse_fun_style
      ~scheme:[ "sched_default", [ `Req "prio" ];
		"sched_rr", [ `Req "prio" ] ]
      s in
  match schedname with
    | "sched_default" ->
	let prio = List.assoc "prio" params in
	let prio_d =
	  try int_of_string prio
	  with _ -> failwith("Parameter prio not a number in: " ^ s) in
	if prio_d < -20 || prio_d > 19 then
	  failwith ("Parameter prio out of range (-20 to 19): " ^ s);
	`Default { sched_def_prio = Some prio_d; 
		   sched_def_affinity = None }
    | "sched_rr" ->
 	let prio = List.assoc "prio" params in
	let prio_d =
	  try int_of_string prio
	  with _ -> failwith("Parameter prio not a number in: " ^ s) in
	if prio_d < 1 || prio_d > 99 then
	  failwith ("Parameter prio out of range (1 to 99): " ^ s);
	`RoundRobin { sched_rr_prio = prio_d; 
		      sched_rr_affinity = None }
    | _ -> assert false
;;


let print_scheduler sched =
  match sched with
    | `Default s ->
	( match s.sched_def_prio with
	    | None -> ""
	    | Some p -> 
		sprintf "sched_default(prio=%d)" p
	)
    | `RoundRobin s ->
	sprintf "sched_rr(prio=%d)" s.sched_rr_prio
;;


let parse_pool s =
  (* Syntax: mac_pool(first_mac_address = XX:XX:XX:XX:XX:XX, number = <n>) *)
  let (schedname, params) =
    parse_fun_style
      ~scheme:[ "mac_pool", [ `Req "first_mac_address"; `Req "number" ] ]
      s in
  let first_mac_addr_s = List.assoc "first_mac_address" params in
  let first_mac_addr =
    try parse_hwaddr first_mac_addr_s
    with _ ->
      failwith("Parameter 'first_mac_address' has invalid format, in: " ^ s) in
  let number_s = List.assoc "number" params in
  let number = 
    try 
      let n = int_of_string number_s in
      if n < 1 then raise Not_found;
      n
    with _ ->
      failwith("Parmeter 'number' is not a valid number, in: " ^ s) in
  (first_mac_addr, number)
;;


let default_pool() =
  (parse_hwaddr "0A:12:BC:AF:32:00", 256)
    (* This is just a random address from the range of locally administered
     * MAC addresses. This range is:
     * X2:XX:XX:XX:XX:XX, X6:XX:XX:XX:XX:XX, XA:XX:XX:XX:XX:XX, 
     * XE:XX:XX:XX:XX:XX.
     *)
;;


let global_config = lazy (
  try
    check_etc_umlmon();
    let ini = new Inifiles.inifile "/etc/umlmon" in
    let vardir = 
      try ini # getval "GLOBAL" "vardir"
      with Inifiles.Invalid_element _ -> "/var/lib/umlmon" in
    { umlvardir = vardir;
      umltcp = ( try ini # getval "GLOBAL" "tcp" = "true"
		 with Inifiles.Invalid_element _ -> false );
      umltcp_redir = ( try ini # getval "GLOBAL" "tcp_redir" = "true"
		 with Inifiles.Invalid_element _ -> false );
      umlmasterpw = ( try Some (parse_password (ini # getval "GLOBAL" "password"))
		      with Inifiles.Invalid_element _ -> None);
      umlsitecmd = ( try Some (ini # getval "GLOBAL" "sitecmd")
		     with Inifiles.Invalid_element _ -> None);
      umlhookcmd = ( try Some (ini # getval "GLOBAL" "hookcmd")
		     with Inifiles.Invalid_element _ -> None);
      umlshared = ( try ini # getval "GLOBAL" "shared" 
		    with Inifiles.Invalid_element _ ->
		      Filename.concat vardir "shared" );
      umldevs = List.map parse_device (ini # getaval "GLOBAL" "device");
      umlhwpool = ( try	parse_pool ( ini # getval "GLOBAL" "pool" )
		    with Inifiles.Invalid_element _ -> default_pool() );
    }
  with
    | Inifiles.Invalid_section n ->
	failwith ("File /etc/umlmon: Section not found: " ^ n)
    | Inifiles.Invalid_element n ->
	failwith ("File /etc/umlmon: Required configuration parameter not found: " ^ n)
);;


let read_global_config() =
  Lazy.force global_config
;;


type logdir_conf =
    { cf_logdir : string;
      cf_any : bool;
    }


let absolute_logfile ldcfg param_name filename =
  if filename = "" then failwith ("Parameter " ^ param_name ^ ": Empty log file name");
  if filename.[0] = '/' then (
    if ldcfg.cf_any then
      filename
    else
      let dir = Filename.dirname filename in
      if dir = ldcfg.cf_logdir then
	filename
      else
	failwith ("Parameter " ^ param_name ^ 
		  ": This directory is not permitted")
  )
  else (
    if String.contains filename '/' then
      failwith ("Parameter " ^ param_name ^ 
		": File name must not contain / in the middle");
    Filename.concat ldcfg.cf_logdir filename
  )


let eth_re = Str.regexp "^eth[0-9]+$"

let is_eth_config name =
  Str.string_match eth_re name 0

let comma_re = Str.regexp "[ \t]*,[ \t]*"

let name_value_re = Str.regexp "\\([a-zA-Z0-9_]+\\)=\\(.*\\)"

let parse_eth_config name value =
  try
    let schemename, params =
      parse_fun_style
	~raise_not_recognized:true
	~scheme:[ "host_to_host", [ `Opt "host_if";
				    `Req "proxy_if";
				    `Req "proxy_addr";
				    `Opt "proxy_hwaddr";
				    `Req "guest_addr";
				    `Opt "guest_hwaddr" ];
		  "bridged", [ `Req "bridge_if";
			       `Req "proxy_if";
			       `Opt "proxy_hwaddr";
			       `Opt "guest_hwaddr" ] ] 
	value
    in
    let param n = List.assoc n params in
    let hwparam n =
      try
	let s = param n in  (* or Not_found *)
	if s = "pool" then 
	  `HW_from_pool
	else
	  ( try `HW (parse_hwaddr s)
	    with _ -> 
	      failwith("Invalid MAC address in parameter '" ^ n ^ "', used in: " ^ value)
	  )
      with Not_found -> `Unconfigured
    in
    let def =
      match schemename with
	| "host_to_host" ->
	    `Host_to_host 
	      { hth_host_if = 
		  (try Some(param "host_if") with Not_found -> None);
		hth_proxy_if = param "proxy_if";
		hth_proxy_addr = param "proxy_addr";
		hth_proxy_hwaddr = hwparam "proxy_hwaddr";
		hth_guest_addr = param "guest_addr";
		hth_guest_hwaddr = hwparam "guest_hwaddr";
	      }
	| "bridged" ->
	    `Bridged
	      { br_bridge_if = param "bridge_if";
		br_proxy_if = param "proxy_if";
		br_proxy_hwaddr = hwparam "proxy_hwaddr";
		br_guest_hwaddr = hwparam "guest_hwaddr";
	      }
	| _ -> assert false
    in
    { ethname = name;
      ethdef = def;
    }
  with
    | Fun_style_not_recognized ->
	(* Support for old-style syntax: *)
	let value_l = Str.split comma_re value in
	let value_al = List.map (fun s -> 
				   if Str.string_match name_value_re s 0 then
				     (Str.matched_group 1 s, Str.matched_group 2 s)
				   else
				     ("", s)) value_l in
	let def =
	  match value_al with
	    | [ "", "host_to_host";
		host_if, proxy_addr;
		proxy_if, guest_addr ] ->
		`Host_to_host 
		  { hth_host_if = Some host_if; 
		    hth_proxy_if = proxy_if;
		    hth_proxy_addr = proxy_addr;
		    hth_proxy_hwaddr = `Unconfigured;
		    hth_guest_addr = guest_addr;
		    hth_guest_hwaddr = `Unconfigured;
		  }
	    | _ ->
		`Unparsed value in
	{ ethname = name;
	  ethdef = def;
	}
;;


let string_of_eth eth = 
  match eth.ethdef with
    | `Host_to_host hth ->
	sprintf
	  "host_to_host(proxy_if=%s, proxy_addr=%s%s, guest_addr=%s%s%s)"
	  hth.hth_proxy_if
	  hth.hth_proxy_addr
	  ( match hth.hth_proxy_hwaddr with
	      | `Unconfigured -> ""
	      | `HW_from_pool -> ", proxy_hwaddr=pool"
	      | `HW hwaddr -> ", proxy_hwaddr=" ^ format_hwaddr hwaddr )
	  hth.hth_guest_addr
	  ( match hth.hth_guest_hwaddr with
	      | `Unconfigured -> ""
	      | `HW_from_pool -> ", guest_hwaddr=pool"
	      | `HW hwaddr -> ", guest_hwaddr=" ^ format_hwaddr hwaddr )
	  ( match hth.hth_host_if with
	      | Some d -> ", host_if=" ^ d
	      | None -> ""
	  )
    | `Bridged br ->
	sprintf
	  "bridged(bridge_if=%s, proxy_if=%s%s%s)"
	  br.br_bridge_if
	  br.br_proxy_if
	  ( match br.br_proxy_hwaddr with
	      | `Unconfigured -> ""
	      | `HW_from_pool -> ", proxy_hwaddr=pool"
	      | `HW hwaddr -> ", proxy_hwaddr=" ^ format_hwaddr hwaddr )
	  ( match br.br_guest_hwaddr with
	      | `Unconfigured -> ""
	      | `HW_from_pool -> ", guest_hwaddr=pool"
	      | `HW hwaddr -> ", guest_hwaddr=" ^ format_hwaddr hwaddr )
    | `Unparsed s ->
	s 
;;


let param_of_eth eth =
  let v = string_of_eth eth in
  { param_name = eth.ethname; param_value = v }
;;


let con_re = Str.regexp "^con[0-9]+$"
let ssl_re = Str.regexp "^ssl[0-9]+$"

let is_con_config name =
  Str.string_match con_re name 0

let is_ssl_config name =
  Str.string_match ssl_re name 0

let pty_re = Str.regexp "pty\\(:\\(.*\\)\\)?$" ;;

let parse_term_config ldcfg name value =
  if Str.string_match pty_re value 0 then (
    let pty =
      try 
	`Pty (Some (absolute_logfile ldcfg name (Str.matched_group 2 value)))
      with
	  Not_found -> `Pty None in
    { termname = name; termdef = pty }
  )
  else
    failwith ("Bad configuration parameter: " ^ name)


let string_of_term ?(trans = fun s -> s) ut =
  match ut.termdef with
    | `Pty None ->
	"pty"
    | `Pty (Some name) ->
	"pty:" ^ trans name 

let param_of_term ut =
  let v = string_of_term ut in
  { param_name = ut.termname; param_value = v }



let ubd_re = Str.regexp "ubd[0-9]+$" ;;

let ws_re = Str.regexp "[ \t\r\n]+";;


let is_ubd_config name =
  Str.string_match ubd_re name 0

let parse_ubd_config name value =
  { diskname = name; diskdef = Jailfile value; disksync = false }

let parse_start_config value =
  match value with
    | "boot" -> `Boot
    | "manual" -> `Manual
    | "off" -> `Off
    | _ -> failwith ("Bad value for parameter 'start'")


let read_vm_config umlid =
  let g = read_global_config() in
  let user_file = 
    Filename.concat (Filename.concat g.umlvardir umlid) "config" in

  let file, ini_is_sys_ini =
    if Sys.file_exists user_file then
      user_file, false
    else
      "/etc/umlmon", true in
  let sys_ini = new Inifiles.inifile "/etc/umlmon" in
  let ini = new Inifiles.inifile file in

  (* First read protected parameters from [sys_ini]: *)
  let (vmuser, vmgroup_opt, sched, jaildir_opt, logdir_opt, mem_limit, eths) =
    try
      let eths = ref [] in
      let vmuser = sys_ini # getval umlid "vmuser" in
      (* The check that the user exists is performed by the call of
       * user_privileges below.
       *)
      let vmgroup_opt =
	try Some (sys_ini # getval umlid "vmgroup")
	with Inifiles.Invalid_element _ -> None in
      (* The check that the group exists and is ok for the user is performed
       * by the call of user_privileges below.
       *)
      let sched =
	try parse_scheduler(sys_ini # getval umlid "scheduler")
	with Inifiles.Invalid_element _ -> 
	  `Default { sched_def_prio = None; sched_def_affinity = None } in
      let jaildir_opt =
	try
	  let jaildir = sys_ini # getval umlid "jaildir" in
	  if jaildir = "" then failwith "Parameter jaildir: is empty";
	  if jaildir.[0] <> '/' then failwith "Parameter jaildir: must be absolute";
	  Some jaildir
	with Inifiles.Invalid_element _ -> None in
      let logdir_opt = 
	try 
	  let logdir = sys_ini # getval umlid "logdir" in
	  if logdir = "" then failwith "Parameter logdir: is empty";
	  if logdir.[0] <> '/' then failwith "Parameter logdir: must be absolute";
	  Some logdir
	with Inifiles.Invalid_element _ -> None in
      let mem_limit = 
	try int_of_string (sys_ini # getval umlid "mem_limit") 
	with Inifiles.Invalid_element _ -> max_int in
      sys_ini # iter
	(fun name value ->
	   if is_eth_config name then (
	     eths := parse_eth_config name value :: !eths
	   )
	)
	umlid;
      (vmuser, vmgroup_opt, sched, jaildir_opt, logdir_opt, mem_limit, !eths)
    with
    | Inifiles.Invalid_section n ->
	failwith ("File /etc/umlmon: Section not found: " ^ n)
    | Inifiles.Invalid_element n ->
	failwith ("File /etc/umlmon: Required configuration parameter not found: " ^ n)
    | Failure msg ->
	failwith ("File " ^ file ^ ": " ^ msg)
  in

  let jaildir =
    match jaildir_opt with
      | None -> 
	  Filename.concat (Filename.concat g.umlvardir umlid) "jail"
      | Some d -> 
	  d
  in
  let logdir =
    match logdir_opt with
      | None -> 
	  Filename.concat (Filename.concat g.umlvardir umlid) "log"
      | Some d -> 
	  d
  in

  let ldcfg = { cf_logdir = logdir; cf_any = false } in

  (* Now read unprotected parameters from [ini]: *)
  try
    let terms = ref [] in
    let disks = ref [] in

    ini # iter
      (fun name value ->
	 if is_con_config name || is_ssl_config name then (
	   terms := parse_term_config ldcfg name value :: !terms
	 )
	 else 
	   if is_ubd_config name then (
	     disks := parse_ubd_config name value :: !disks
	   )
      )
      umlid;

    let spec = 
      { umlid = umlid;
	umlvmuser = vmuser;
	umlvmuserpw = ( try Some(parse_password (ini # getval umlid "password"))
			with Inifiles.Invalid_element _ -> None );
	umlvmgroup = vmgroup_opt;
	umlsched = sched;
	umlpriv = Umlmon_common.user_privileges vmuser vmgroup_opt;
	umljaildir = jaildir;
	umllogdir = logdir;
	umlmem = int_of_string (ini # getval umlid "mem");
	umlmemlimit = mem_limit;
	umlkernel = ini # getval umlid "kernel";
	umlkernelargs = ini # getaval umlid "kernelarg";
	umlterms = List.rev !terms;
	umldisks = List.rev !disks;
	umleths = List.rev eths;
	umlsync = ( try
		      let v = ini # getval umlid "sync_disks" in
		      Str.split ws_re v
		    with Inifiles.Invalid_element _ -> default_sync_disks );
	umlmonlog = ( try 
			Some 
			  (absolute_logfile ldcfg "logfile" (ini # getval umlid "logfile"))
		      with Inifiles.Invalid_element _ -> None );
	umlstart = parse_start_config
		     (try ini#getval umlid "start"
		      with Inifiles.Invalid_element _ -> "manual");
      }  in
    
    spec
  with
    | Inifiles.Invalid_section n ->
	failwith ("File " ^ file ^ ": Section not found: " ^ n)
    | Inifiles.Invalid_element n ->
	failwith ("File " ^ file ^ ": Required configuration parameter not found: " ^ n)
    | Failure msg ->
	failwith ("File " ^ file ^ ": " ^ msg)
;;


let string_of_start =
  function
    | `Boot -> "boot"
    | `Manual -> "manual"
    | `Off -> "off"


let get_inifile ?(priv=false) umlid spec =
  let g = read_global_config() in

  let unabsolute_logfile filename =
    let d = Filename.dirname filename in
    if d = spec.umllogdir then
      Filename.basename filename
    else
      d in

  let ini = new Inifiles.inifile "/dev/null" in

  if priv then (
    ini # setval umlid "vmuser" spec.umlvmuser;
    ( match spec.umlvmgroup with
	| None -> ()
	| Some g -> ini # setval umlid "vmgroup" g
    );
    ini # setval umlid "jaildir" spec.umljaildir;
    ini # setval umlid "logdir" spec.umllogdir;
    ini # setval umlid "mem_limit" (string_of_int spec.umlmemlimit);
  );

  ( match spec.umlvmuserpw with
      | None -> ()
      | Some pw -> ini # setval umlid "password" (print_password pw)
  );
  ini # setval umlid "start" (string_of_start spec.umlstart);
  ini # setval umlid "mem" (string_of_int spec.umlmem);
  ini # setval umlid "kernel" spec.umlkernel;
  List.iter
    (fun v ->
       ini # setval umlid "kernelarg" v
    )
    spec.umlkernelargs;
  ( match spec.umlmonlog with
      | None -> ()
      | Some f -> ini # setval umlid "logfile" (unabsolute_logfile f)
  );
  List.iter
    (fun t ->
       let v = string_of_term ~trans:unabsolute_logfile t in
       ini # setval umlid t.termname v
    )
    spec.umlterms;
  List.iter
    (fun d ->
       let Jailfile v = d.diskdef in
       ini # setval umlid d.diskname v
    )
    spec.umldisks;
  if priv then (
    List.iter
      (fun e ->
	 let v = string_of_eth e in
	 ini # setval umlid e.ethname v
      )
      spec.umleths
  );
  ini # setval umlid "sync_disks" (String.concat " " spec.umlsync);

  ini
;;

let write_vm_config umlid spec =
  let ini = get_inifile umlid spec in
  let g = read_global_config() in
  let user_file = 
    Filename.concat (Filename.concat g.umlvardir umlid) "config" in
  ini # save ~file:user_file ()
;;


let vm_list() =
  let sys_ini = new Inifiles.inifile "/etc/umlmon" in
  List.filter
    (fun s -> s <> "GLOBAL" && s <> "LICENSE")
    sys_ini # sects
;;


let con_re = Str.regexp "^con[0-9]+$";;
let ssl_re = Str.regexp "^ssl[0-9]+$";;


let proc_config_get gs =
  protect
    (fun session () reply ->
       Umlmon_auth.ensure_authenticated session;
       if gs.umlvm = `Inactive then (
	 (* Re-read configuration *)
	 let spec = read_vm_config (Lazy.force gs.umlspec).umlid in (* or fail! *)
	 gs.umlspec <- lazy spec;
       );

       let spec = Lazy.force gs.umlspec in
       let con_terms =
	 List.filter
	   (fun ut -> Str.string_match con_re ut.termname 0) 
	   spec.umlterms in
       let ssl_terms =
	 List.filter
	   (fun ut -> Str.string_match ssl_re ut.termname 0) 
	   spec.umlterms in
       
       let r =
	 { cfg_privsetup =
	     if rpc_over_inet session then
	       [| { param_name = "mem_limit"; 
		    param_value = string_of_int spec.umlmemlimit };
	       |]
	     else 
	       [| { param_name = "vmuser"; param_value = spec.umlvmuser };
		  { param_name = "vmgroup"; 
		    param_value = 
		      ( match spec.umlvmgroup with
			  | None -> "" | Some g -> g ) };
	          { param_name = "scheduler"; 
		    param_value = print_scheduler spec.umlsched };
		  { param_name = "jaildir"; param_value = spec.umljaildir };
		  { param_name = "logdir"; param_value = spec.umllogdir };
		  { param_name = "mem_limit"; 
		    param_value = string_of_int spec.umlmemlimit };
	       |];
	   cfg_start = { param_name = "start"; param_value = string_of_start spec.umlstart };
	   cfg_start_rw = true;
	   cfg_logfile =
	     ( match spec.umlmonlog with
		 | None -> 
		     None
		 | Some log ->
		     Some { param_name = "logfile"; param_value = log } );
	   cfg_logfile_rw = true;
	   cfg_kernel = 
	     { param_name = "kernel"; param_value = spec.umlkernel };
	   cfg_kernel_rw = true;
	   cfg_kernelargs =
	     Array.of_list
	       ( List.map
		   (fun arg -> { param_name = "kernelarg"; param_value = arg })
		   spec.umlkernelargs );
	   cfg_kernelargs_rw = true;
	   cfg_mem = 
	     { param_name = "mem"; param_value = string_of_int spec.umlmem };
	   cfg_mem_rw = true;
	   cfg_mem_hot = false;
	   cfg_con = Array.of_list ( List.map param_of_term con_terms );
	   cfg_con_rw = true;
	   cfg_con_hot = false;
	   cfg_ssl = Array.of_list ( List.map param_of_term ssl_terms );
	   cfg_ssl_rw = true;
	   cfg_ssl_hot = false;
	   cfg_ubd = 
	     Array.of_list
	       ( List.map
		   (fun ud ->
		      let Jailfile v = ud.diskdef in
		      { param_name = ud.diskname; param_value = v })
		   spec.umldisks );
	   cfg_ubd_rw = true;
	   cfg_ubd_hot = false;
	   cfg_eth = Array.of_list ( List.map param_of_eth spec.umleths );
	   cfg_eth_rw = false;
	   cfg_eth_hot = false;
	   cfg_extension = [|
	     { param_name = "sync_disks";
	       param_value = String.concat " " spec.umlsync }
	   |];
	 } in
       reply r
    )
;;


let proc_config_set gs =
  protect
    (fun session cfg reply ->
       Umlmon_auth.ensure_authenticated session;
       try
	 let spec = Lazy.force gs.umlspec in

	 if gs.umlvm <> `Inactive then 
	   failwith "VM must be inactive to set configuration";

	 if cfg.cfg_privsetup <> [| |] then
	   failwith "It is not permitted to set privileged parameters";

	 let ldcfg = { cf_logdir = spec.umllogdir; cf_any = false } in

	 let spec' = 
	   { spec with
	       umlstart = ( let p = cfg.cfg_start in
			  if p.param_name <> "start" then
			     failwith (p.param_name ^ ": bad name");
			  parse_start_config p.param_value );

	       umlmonlog = ( match cfg.cfg_logfile with
			       | None -> None
			       | Some p ->
				   if p.param_name <> "logfile" then
				     failwith (p.param_name ^ ": bad name");
				   let v = p.param_value in
				   let v' =
				     absolute_logfile ldcfg "logfile" v in
				   Some v');
	       umlkernel = ( let p = cfg.cfg_kernel in
			   if p.param_name <> "kernel" then
			     failwith (p.param_name ^ ": bad name");
			   p.param_value );
	       umlkernelargs = ( List.map
				(fun p ->
				   if p.param_name <> "kernelarg" then
				     failwith (p.param_name ^ ": bad name");
				   p.param_value)
				(Array.to_list cfg.cfg_kernelargs) );
	       umlmem = ( let p = cfg.cfg_mem in
			if p.param_name <> "mem" then
			  failwith (p.param_name ^ ": bad name");
			int_of_string p.param_value );
	       umlterms = ( List.map
			    (fun p ->
			       if not (is_con_config p.param_name) then
				 failwith (p.param_name ^ ": bad name");
			       parse_term_config ldcfg p.param_name p.param_value)
			    (Array.to_list cfg.cfg_con) @
			  List.map
			    (fun p ->
			       if not (is_ssl_config p.param_name) then
				 failwith (p.param_name ^ ": bad name");
			       parse_term_config ldcfg p.param_name p.param_value)
			    (Array.to_list cfg.cfg_ssl) );
	       umldisks = ( List.map
			    (fun p ->
			       if not (is_ubd_config p.param_name) then
				 failwith (p.param_name ^ ": bad name");
			       parse_ubd_config p.param_name p.param_value)
			    (Array.to_list cfg.cfg_ubd) );
	       umleths = ( List.map
			    (fun p ->
			       if not (is_eth_config p.param_name) then
				 failwith (p.param_name ^ ": bad name");
			       parse_eth_config p.param_name p.param_value)
			    (Array.to_list cfg.cfg_eth) );
	       umlsync = ( try
			     let p = 
			       List.find 
				 (fun p -> p.param_name = "sync_disks") 
				 (Array.to_list cfg.cfg_extension) in
			     Str.split ws_re p.param_value
			   with 
			       Not_found -> default_sync_disks
			 )
	   } in
	 (* TODO: Check for duplicate parameters! *)
	 write_vm_config spec'.umlid spec';
	 let spec'' = read_vm_config spec'.umlid in (* or fail! *)
	 gs.umlspec <- lazy spec'';
	 reply `code_ok
       with
	 | Failure msg ->
	     reply (`code_error msg);
    )
;;


let proc_config_hot_add gs =
  protect
    (fun session cfg reply ->
       Umlmon_auth.ensure_authenticated session;
       reply (`code_error "Procedure not implemented")
    )
;;


let proc_config_hot_remove gs =
  protect
    (fun session cfg reply ->
       Umlmon_auth.ensure_authenticated session;
       reply (`code_error "Procedure not implemented")
    )
;;

